require 'expo2'
require 'hash_helper'

class Expo
  CALL_ORDER = [:aside, :expose, :sub_expo, :collection]
  TRIVIAL_PRESENTER = Class.new do
    def present(object)
      object
    end
  end.new

  def initialize(presenter=nil, &block)
    @presenter = presenter
    @block = block
    @message_map = {}
    @subs = {}
    @collections = {}
    @inner_calls = []
    @id = self.object_id
  end

  def expo_collection(objects, context={})
    objects.map.with_index do |object, index|
      Expo.new(@presenter, &@block).expo(object, context.merge({index: index}))
    end
  end

  def reset
    @message_map = {}
    @subs = {}
    @collections = {}
    @inner_calls = []
  end

  def expo(object, context={})

    @object = object
    @context = context
    if object.nil?
      return nil
    end
    @preso = Proxy.new(
      object,
      (@presenter || TRIVIAL_PRESENTER).present(object)
    )
    if !@block
      return {}
    end
    block_return = self.instance_exec @object, context, &@block
    @inner_calls.each do |call|
      self.send(call[:method], *call[:args], &call[:block])
    end
    exposition = {}
    @message_map.each do |k,v|
      exposition[v] = @preso.send(k)
    end
    exposition.merge!(@subs)
    exposition.merge!(@collections)
    self.reset
    if block_return
      exposition.merge(block_return)
    else
      exposition
    end
  end

  def aside(*messages)
    array, hash = parse_arguments(messages)
    @context.merge!(hasherize(array) {|msg| @preso.send(msg) })
    hash.each do |k,v|
      @context[v] = @preso.send(k)
    end
    nil
  end

  def expose(*messages)
    array, hash = parse_arguments(messages)
    @message_map.merge!(hasherize(array).merge(hash))
    nil
  end

  def sub_expo(key, object, sub_expo_passed)
    @inner_calls << {
      method: :sub_expo_inner,
      args: [key, object, sub_expo_passed],
      block: nil
    }
    nil
  end

  def collection(key, collection, item_expo)
    @inner_calls << {
      method: :collection_inner,
      args: [key, collection, item_expo],
      block: nil
    }
    nil
  end

  private

  def set_id(id)
    @id = id
    self
  end

  def get_id
    @id
  end

  def sub_expo_inner(key, object, sub_expo_passed)
    recurse_expo = sub_expo_passed
    sub_id = sub_expo_passed.send(:get_id)
    if sub_id == @id
      recurse_expo = Expo.new(@presenter, &@block).send(:set_id, @id)
    end
    sub = recurse_expo.expo(object, @context)
    @subs[key] = sub if sub
    nil
  end

  def collection_inner(key, collection, item_expo)
    recurse_expo = item_expo
    sub_id = item_expo.send(:get_id)
    if sub_id == @id
      recurse_expo = Expo.new(@presenter, &@block).send(:set_id, @id)
    end
    if !collection
      @collections[key] = nil
      return nil
    end
    collection_exposition = []
    collection.each.with_index do |item, index|
      collection_exposition << recurse_expo.expo(item, @context.merge({
        "#{key}_index".to_sym => index
      }))
    end
    @collections[key] = collection_exposition
    nil
  end

  def hasherize(array, &block)
    block ||= ->(e) { e }
    Hash[*(array.map{|e| [e, block.call(e)] }.flatten(1))]
  end

  def parse_arguments(args)
    last = args.pop
    array = args
    hash = {}
    if last.is_a?(Hash)
      hash = last
    elsif last
      array << last
    end
    [array, hash]
  end

  class Proxy
    def initialize(*objects)
      @objects = objects
    end

    def method_missing(method, *args, &block)
      @objects.detect{|obj| obj.respond_to? method}.send(
        method, *args, &block
      )
    end

  end
end
