require 'null_expo'
require 'hash_helper'
require 'argument_parser'
require 'presenters'

class Expo2
  def initialize(prototype=nil, &block)
    @block = block
    @prototype = prototype || NullExpo.new
    @presenters ||= []
    self.instance_exec &@block if @block
  end

  def inspect
    "<Expo2:#{self.object_id}, @prototype=#{@prototype.inspect}, @presenters=#{@presenters.map(&:inspect)}>"
  end


  def expo(object=nil, context={}, index=nil)
    preso = @prototype.expo(object, context, index)
    @presenters.each do |pres|
      #TODO: warn on key collision
      preso = HashHelper.new.deep_merge(preso, pres.go(object, context, index))
    end
    preso
  end

  def expo_collection(objects, context={})
    objects.map.with_index{|o, i| expo(o, context, i) }
  end

  private

  def expose(*keys)
    array, hash = ArgumentParser.new.parse_arguments(keys)
    key_map = HashHelper.new.hasherize(array).merge(hash)
    @presenters << Exposer.new(key_map)
  end

  def aside(presenter, *keys)
    array, hash = ArgumentParser.new.parse_arguments(keys)
    key_map = HashHelper.new.hasherize(array).merge(hash)
    @presenters << Asider.new(presenter, key_map)
  end

  def augment(presenter, *keys)
    array, hash = ArgumentParser.new.parse_arguments(keys)
    key_map = HashHelper.new.hasherize(array).merge(hash)
    @presenters << Pres.new(presenter, key_map)
  end

  def merge(&block)
    @presenters << Merger.new(block)
  end

  def merge_expo(pres, sub_expo=nil, &block)
    @presenters << MergeExpo.new(pres, sub_expo, block)
  end

  def sub_expo(pres, key, sub_expo=nil, &block)
    @presenters << SubExpo.new(pres, key, sub_expo, block)
  end

  def recurse(pres, key)
    @presenters << Recurser.new(pres, key, self)
  end

  def collection(associator, key, expo=nil, &block)
    @presenters << SubCollection.new(associator, key, expo, block)
  end

  def inline(&block)
    @presenters << InlineExpo.new(block)
  end

end
