class Expo3
  class Builder
    def initialize(expo)
      @expo = expo
      @output = {}
      @blocks = []
    end

    def expose(*keys)
      key_map = key_map(keys)
      output = @output
      @blocks << Proc.new do |obj, **context|
        key_map.each do |k,v|
          output[v] = obj.send(k)
        end
      end
    end

    def augment(another, *keys)
      return if another.nil?
      key_map = key_map(keys)
      output = @output
      @blocks << Proc.new do |obj, **context|
        key_map.each do |k,v|
          output[v] = another.send(k)
        end
      end
    end

    def sub_expo(key, another, sub_expo=nil, &block)
      return if another.nil?
      output = @output
      @blocks << Proc.new do |obj, **context|
        sub = sub_expo || Expo3.new(&block)
        out = sub.expo(another, **context)
        output[key] = out
      end
    end

    def merge_expo(another, sub_expo=nil, &block)
      return if another.nil?
      output = @output
      @blocks << Proc.new do |obj, **context|
        sub = sub_expo || Expo3.new(&block)
        out = sub.expo(another,**context)
        output.merge! out
      end
    end

    def collection(key, anothers, sub_expo=nil, &block)
      return if anothers.nil?
      output = @output
      sub = (sub_expo || Expo3.new).upgrade(&block)
      @blocks << Proc.new do |obj,**context|
        out = anothers.map.with_index do |o, i|
          if sub.expects?(:index)
            sub.expo(o, **(context.merge(index: i)))
          else
            sub.expo(o, **context)
          end
        end
        output[key] = out
      end
      sub
    end

    def inline(hash)
      @output.merge!(hash)
    end

    def current_expo
      @expo
    end

    def output(obj=nil,**context)
      @blocks.each do |b|
        b.call(obj,**context)
      end
      @output
    end

    private
    def key_map(keys)
      array, hash = ArgumentParser.new.parse_arguments(keys)
      HashHelper.new.hasherize(array).merge(hash)
    end
  end

  def initialize(**context, &block)
    @block = block
    @context = context
  end

  def expects?(keyparam)
    @block.parameters.detect{|k,param| param == keyparam }
  end

  def fix(**context)
    Expo3.new(**(@context.merge(context)), &@block)
  end

  def upgrade(&block)
    return self unless block
    return Expo3.new(**(@context), &block) unless @block
    this = self
    Expo3.new do |obj|
      merge_expo obj, this
      merge_expo obj, Expo3.new(&block)
    end
  end

  def expo(obj=nil, **context)
    builder = Builder.new(self)
    builder.instance_exec(obj, **(@context.merge(context)), &@block) if @block
    builder.output(obj, **context)
  end

  def expo_collection(coll)
    coll.map(&method(:expo))
  end

end
